from typing import TYPE_CHECKING

import pytest
from sqlalchemy.orm import Session
from uuid_utils.compat import UUID

if TYPE_CHECKING:
    from mixemy.models import IdAuditModel


@pytest.mark.database
@pytest.mark.integration
def test_main(
    session: Session, init_db: None, item_model: "type[IdAuditModel]"
) -> None:
    """Test the main functionality of the ItemService class.

    This test covers the following scenarios:
    1. Creating items.
    2. Reading items by ID.
    3. Updating an item.
    4. Reading multiple items with filters.
    5. Deleting items.

    Args:
        session (Session): The database session to use for the test.
        init_db (None): Fixture to initialize the database.
        item_model (type[IdAuditModel]): The model class to use for the items.

    Raises:
        - The created items have the correct values.
        - The read items are not None and have the correct values.
        - The updated item has the correct updated values.
        - The correct number of items are returned when filtering.
        - The deleted items are no longer present in the database.
    """
    from mixemy import repositories, schemas, services

    ItemModel = item_model

    class ItemInput(schemas.InputSchema):
        value: str

    class ItemUpdate(ItemInput):
        nullable_value: str | None

    class ItemFilter(schemas.InputSchema):
        value: list[str]

    class ItemOutput(schemas.IdAuditOutputSchema):
        value: str
        nullable_value: str | None

    class ItemRepository(repositories.BaseSyncRepository[ItemModel]):
        model = ItemModel

    class ItemService(services.BaseSyncService[ItemRepository, ItemOutput]):
        repository_type = ItemRepository
        output_schema_type = ItemOutput

    item_service = ItemService(db_session=session)

    test_one = ItemInput(value="test_one")
    test_two = ItemInput(value="test_two")
    test_three = ItemInput(value="test_one")
    test_one_update = ItemUpdate(value="test_one", nullable_value="test_one_updated")
    test_one_id = None

    item_one = item_service.create(object_in=test_one)
    item_two = item_service.create(object_in=test_two)
    item_service.create(object_in=test_three)

    test_one_id = item_one.id

    assert item_one.value == "test_one"
    assert item_two.value == "test_two"

    item_one = item_service.read(id=item_one.id)
    item_two = item_service.read(id=item_two.id)

    assert item_one is not None
    assert item_two is not None
    assert item_one.value == "test_one"
    assert item_one.nullable_value is None
    assert item_two.value == "test_two"
    assert item_two.nullable_value is None

    item_one = item_service.update(id=item_one.id, object_in=test_one_update)

    assert item_one is not None
    assert item_one.value == "test_one"
    assert item_one.nullable_value == "test_one_updated"
    assert item_one.id == test_one_id

    items = item_service.read_multiple(filters=ItemFilter(value=["test_one"]))

    assert len(items) == 2

    item_service.delete(id=item_one.id)

    item_one = item_service.read(id=item_one.id)
    item_two = item_service.read(id=item_two.id)

    assert item_one is None
    assert item_two is not None
    assert item_two.value == "test_two"

    item_service.delete(id=item_two.id)

    item_two = item_service.read(id=item_two.id)

    assert item_two is None


def test_recursive_model(
    session: Session, init_db: None, recursive_item_model: "type[IdAuditModel]"
) -> None:
    """Test the recursive model functionality.

    This test verifies the creation and retrieval of a recursive item model
    using the provided session and database initialization.

    Args:
        session (Session): The database session to use for the test.
        init_db (None): A fixture to initialize the database.
        recursive_item_model (type[IdAuditModel]): The recursive item model type.

    Raises:
        The created item is not None.
        The created item has two sub-items with the expected values.
        The created item has a singular sub-item with the expected value.
    """
    from mixemy import repositories, schemas, services

    RecursiveItemModel = recursive_item_model

    class SubItemInput(schemas.InputSchema):
        value: str

    class SingularSubItemInput(schemas.InputSchema):
        value: str

    class SubItemOutput(schemas.IdAuditOutputSchema):
        value: str

    class SingularSubItemOutput(schemas.IdAuditOutputSchema):
        value: str

    class ItemInput(schemas.InputSchema):
        sub_items: list[SubItemInput]
        singular_sub_item: SingularSubItemInput

    class ItemOutput(schemas.IdAuditOutputSchema):
        sub_items: list[SubItemOutput]
        singular_sub_item: SingularSubItemOutput

    class ItemRepository(repositories.BaseSyncRepository[RecursiveItemModel]):
        model = RecursiveItemModel

    class ItemService(
        services.BaseSyncService[
            ItemRepository,
            ItemOutput,
        ]
    ):
        repository_type = ItemRepository
        output_schema_type = ItemOutput
        default_recursive_model_conversion = True

    item_service = ItemService(db_session=session)

    test_item = ItemInput(
        sub_items=[
            SubItemInput(value="sub_item_one"),
            SubItemInput(value="sub_item_two"),
        ],
        singular_sub_item=SingularSubItemInput(value="singular_sub_item"),
    )

    item_id = item_service.create(object_in=test_item).id

    item = item_service.read(id=item_id)

    assert item is not None
    assert len(item.sub_items) == 2
    assert item.sub_items[0].value == "sub_item_one"
    assert item.sub_items[1].value == "sub_item_two"
    assert item.singular_sub_item.value == "singular_sub_item"


def test_permission_model(
    session: Session,
    init_db: None,
    permission_item_model: "type[IdAuditModel]",
    user_model: "type[IdAuditModel]",
) -> None:
    """Test the permission model functionality.

    This test verifies the creation and retrieval of permission items using the provided session
    and database initialization.

    Args:
        session (Session): The database session to use for the test.
        init_db (None): A fixture to initialize the database.
        permission_item_model (type[IdAuditModel]): The permission item model type.
        user_model (type[IdAuditModel]): The user model type.
    """
    from mixemy import exceptions, repositories, schemas, services

    PermissionItemModel = permission_item_model

    class ItemInput(schemas.InputSchema):
        value: str
        user_id: UUID

    class ItemOutput(schemas.IdAuditOutputSchema):
        value: str

    class UserRepository(repositories.BaseSyncRepository[user_model]):
        model = user_model

    class UserService(
        services.BaseSyncService[UserRepository, schemas.IdAuditOutputSchema]
    ):
        repository_type = UserRepository
        output_schema_type = schemas.IdAuditOutputSchema

    class ItemRepository(repositories.PermissionSyncRepository[PermissionItemModel]):
        model = PermissionItemModel

    class ItemService(services.PermissionSyncService[ItemRepository, ItemOutput]):
        repository_type = ItemRepository
        output_schema_type = ItemOutput

    item_service = ItemService(db_session=session)
    user_service = UserService(db_session=session)

    user_one = user_service.create(object_in=schemas.InputSchema())
    user_two = user_service.create(object_in=schemas.InputSchema())

    test_one = ItemInput(value="test_one", user_id=user_one.id)
    test_two = ItemInput(value="test_two", user_id=user_one.id)
    test_three = ItemInput(value="test_three", user_id=user_two.id)

    created_item_one = item_service.create(object_in=test_one)
    item_service.create(object_in=test_two)
    item_service.create(object_in=test_three)

    assert created_item_one.value == "test_one"

    all_created_items = item_service.read_multiple_with_permission(user_id=user_one.id)

    assert len(all_created_items) == 2

    with pytest.raises(exceptions.MixemyRepositoryPermissionError):
        item_service.delete_with_permission(
            id=created_item_one.id, user_id=test_three.user_id
        )

    with pytest.raises(exceptions.MixemyRepositoryPermissionError):
        item_service.update_with_permission(
            id=created_item_one.id, object_in=test_one, user_id=test_three.user_id
        )

    assert (
        item_service.read_with_permission(
            id=created_item_one.id, user_id=test_three.user_id
        )
        is None
    )

    assert (
        item_service.read_with_permission(
            id=created_item_one.id, user_id=test_one.user_id
        )
        == created_item_one
    )

    item_service.delete_with_permission(
        id=created_item_one.id, user_id=test_one.user_id
    )

    assert (
        item_service.read_with_permission(
            id=created_item_one.id, user_id=test_one.user_id
        )
        is None
    )


def test_read_only_repository(
    session: Session, init_db: None, item_model: "type[IdAuditModel]"
) -> None:
    """Test the read-only repository functionality.

    This test verifies that the read-only repository raises the appropriate error
    when attempting to perform write operations.
    """
    from mixemy import exceptions, repositories, schemas, services

    ItemModel = item_model

    class ItemInput(schemas.InputSchema):
        value: str

    class ItemUpdate(ItemInput):
        nullable_value: str | None

    class ItemOutput(schemas.IdAuditOutputSchema):
        value: str
        nullable_value: str | None

    class ItemRepository(repositories.BaseSyncRepository[ItemModel]):
        model = ItemModel

    class ItemService(services.BaseSyncService[ItemRepository, ItemOutput]):
        repository_type = ItemRepository
        output_schema_type = ItemOutput

    item_service = ItemService(db_session=session)

    test_one = ItemInput(value="test_one")
    test_two = ItemInput(value="test_two")
    test_one_update = ItemUpdate(value="test_one", nullable_value="test_one_updated")

    item_one = item_service.create(object_in=test_one)
    item_two = item_service.create(object_in=test_two)

    class NewItemRepository(repositories.BaseSyncRepository[ItemModel]):
        model = ItemModel
        default_is_read_only = True

    class NewItemService(services.BaseSyncService[NewItemRepository, ItemOutput]):
        repository_type = NewItemRepository
        output_schema_type = ItemOutput

    new_item_service = NewItemService(db_session=session)

    item_one = new_item_service.read(id=item_one.id)
    item_two = new_item_service.read(id=item_two.id)

    assert item_one is not None
    assert item_two is not None
    assert item_one.value == "test_one"
    assert item_one.nullable_value is None
    assert item_two.value == "test_two"
    assert item_two.nullable_value is None

    with pytest.raises(exceptions.MixemyRepositoryReadOnlyError):
        new_item_service.update(id=item_one.id, object_in=test_one_update)

    with pytest.raises(exceptions.MixemyRepositoryReadOnlyError):
        new_item_service.create(object_in=test_one)

    with pytest.raises(exceptions.MixemyRepositoryReadOnlyError):
        new_item_service.delete(id=item_one.id)



from typing import Any, override

from sqlalchemy import Select, or_

from mixemy.repositories.permission_strategies._base import PermissionStrategy
from mixemy.types import BaseModelT, SelectT


class LoosePermissionStrategy(PermissionStrategy):
    @override
    def add_permission_filter(
        self,
        statement: Select[SelectT],
        user_id: Any,
    ) -> Select[SelectT]:
        if user_id is None:
            return statement

        if self.user_joined_table is None:
            statement = statement.where(
                or_(
                    getattr(self.model, self.user_id_attribute) == None,  # noqa: E711
                    getattr(self.model, self.user_id_attribute) == user_id,
                )
            )
        else:
            joined_table = getattr(self.model, self.user_joined_table)
            statement = statement.join(joined_table, isouter=True).where(
                getattr(joined_table, self.user_id_attribute) == user_id
            )

        return statement

    @override
    def check_permission_on_db_oject(
        self,
        db_object: BaseModelT | None,  # type: ignore no
        user_id: Any,
    ) -> tuple[bool, Any]:
        if db_object is None or user_id is None:
            return True, None

        object_id = (
            getattr(db_object, self.user_id_attribute)
            if self.user_joined_table is None
            else getattr(
                getattr(db_object, self.user_joined_table), self.user_id_attribute
            )
        )

        return (object_id is None) or (user_id == object_id), object_id



"""This module initializes and exports various types used throughout the mixemy package.

Exports:
    ID (class): A class representing an identifier.
    AuditModelT (type): A type representing an audit model.
    AuditPaginationSchemaT (type): A type representing an audit pagination schema.
    BaseModelT (type): A type representing a base model.
    BaseSchemaT (type): A type representing a base schema.
    CreateSchemaT (type): A type representing a create schema.
    FilterSchemaT (type): A type representing a filter schema.
    IdAuditModelT (type): A type representing an ID audit model.
    IdModelT (type): A type representing an ID model.
    OutputSchemaT (type): A type representing an output schema.
    PaginationSchemaT (type): A type representing a pagination schema.
    RepositoryAsyncT (type): A type representing an asynchronous repository.
    RepositorySyncT (type): A type representing a synchronous repository.
    ResultT (type): A type representing a result.
    SelectT (type): A type representing a select operation.
    SessionType (type): A type representing a session.
    UpdateSchemaT (type): A type representing an update schema.
"""

from ._models import BaseModelT
from ._repositories import SelectT, permission_strategies
from ._schemas import (
    BaseSchemaT,
    CreateSchemaT,
    FilterSchemaT,
    OutputSchemaT,
    UpdateSchemaT,
)
from ._services import (
    PermissionAsyncRepositoryT,
    PermissionSyncRepositoryT,
    RepositoryAsyncT,
    RepositorySyncT,
)

__all__ = [
    "BaseModelT",
    "BaseSchemaT",
    "CreateSchemaT",
    "FilterSchemaT",
    "OutputSchemaT",
    "PermissionAsyncRepositoryT",
    "PermissionSyncRepositoryT",
    "RepositoryAsyncT",
    "RepositorySyncT",
    "SelectT",
    "UpdateSchemaT",
    "permission_strategies",
]



"""
This module provides functionality for [describe the main purpose of the module].

Classes:
    [ClassName]: [Brief description of the class].

Functions:
    [function_name]: [Brief description of the function].
"""



from asyncio import current_task
from collections.abc import AsyncGenerator, Generator
from uuid import UUID

import pytest
import pytest_asyncio
from sqlalchemy import UUID as SQLAUUID
from sqlalchemy import Engine, ForeignKey, String, create_engine, make_url
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    async_scoped_session,
    async_sessionmaker,
    create_async_engine,
)
from sqlalchemy.orm import (
    Mapped,
    Session,
    mapped_column,
    relationship,
    scoped_session,
    sessionmaker,
)
from testcontainers.postgres import PostgresContainer

from mixemy import models


class AsyncItemModel(models.IdAuditModel):
    __table_args__ = {"extend_existing": True}  # noqa: RUF012
    value: Mapped[str] = mapped_column(String)


class ItemModel(models.IdAuditModel):
    __table_args__ = {"extend_existing": True}  # noqa: RUF012
    value: Mapped[str] = mapped_column(String)
    nullable_value: Mapped[str | None] = mapped_column(String, nullable=True)


class RecursiveItemModel(models.IdAuditModel):
    __table_args__ = {"extend_existing": True}  # noqa: RUF012

    sub_items: Mapped[list["SubItemModel"]] = relationship(
        "SubItemModel", back_populates="item"
    )
    singular_sub_item: Mapped["SingularSubItemModel"] = relationship(
        "SingularSubItemModel", back_populates="item"
    )


class SubItemModel(models.IdAuditModel):
    __table_args__ = {"extend_existing": True}  # noqa: RUF012
    value: Mapped[str] = mapped_column(String)
    item_id: Mapped[UUID] = mapped_column(SQLAUUID, ForeignKey("recursive_item.id"))

    item: Mapped["RecursiveItemModel"] = relationship(
        "RecursiveItemModel", back_populates="sub_items"
    )


class SingularSubItemModel(models.IdAuditModel):
    __table_args__ = {"extend_existing": True}  # noqa: RUF012
    value: Mapped[str] = mapped_column(String)
    item_id: Mapped[UUID] = mapped_column(SQLAUUID, ForeignKey("recursive_item.id"))

    item: Mapped["RecursiveItemModel"] = relationship(
        "RecursiveItemModel", back_populates="singular_sub_item"
    )


class UserModel(models.IdAuditModel):
    __table_args__ = {"extend_existing": True}  # noqa: RUF012

    items: Mapped[list["PermissionItemModel"]] = relationship(
        "PermissionItemModel", back_populates="user"
    )


class PermissionItemModel(models.IdAuditModel):
    __table_args__ = {"extend_existing": True}  # noqa: RUF012
    value: Mapped[str] = mapped_column(String)
    user_id: Mapped[UUID] = mapped_column(SQLAUUID, ForeignKey("user.id"))

    user: Mapped["UserModel"] = relationship("UserModel", back_populates="items")


@pytest.fixture(scope="module")
def item_model() -> type[ItemModel]:
    return ItemModel


@pytest.fixture(scope="module")
def async_item_model() -> type[AsyncItemModel]:
    return AsyncItemModel


@pytest.fixture(scope="module")
def recursive_item_model() -> type[RecursiveItemModel]:
    return RecursiveItemModel


@pytest.fixture(scope="module")
def sub_item_model() -> type[SubItemModel]:
    return SubItemModel


@pytest.fixture(scope="module")
def singular_sub_item_model() -> type[SingularSubItemModel]:
    return SingularSubItemModel


@pytest.fixture(scope="module")
def user_model() -> type[UserModel]:
    return UserModel


@pytest.fixture(scope="module")
def permission_item_model() -> type[PermissionItemModel]:
    return PermissionItemModel


@pytest.fixture(scope="module")
def db_container() -> Generator[PostgresContainer, None, None]:
    with PostgresContainer("postgres:latest") as postgres:
        yield postgres


@pytest.fixture(scope="module")
def engine(db_container: PostgresContainer) -> Engine:
    return create_engine(make_url(db_container.get_connection_url()))


@pytest.fixture(scope="module")
def scoped_session_maker(engine: Engine) -> scoped_session[Session]:
    return scoped_session(sessionmaker(bind=engine))


@pytest.fixture(scope="module")
def async_engine(db_container: PostgresContainer) -> AsyncEngine:
    return create_async_engine(
        make_url(db_container.get_connection_url(driver="asyncpg"))
    )


@pytest.fixture(scope="module")
def async_scoped_session_maker(
    async_engine: AsyncEngine,
) -> async_scoped_session[AsyncSession]:
    return async_scoped_session(
        async_sessionmaker(bind=async_engine, expire_on_commit=False),
        scopefunc=current_task,
    )


@pytest.fixture
def session(
    scoped_session_maker: scoped_session[Session],
) -> Generator[Session]:
    session = scoped_session_maker()
    try:
        yield session
    finally:
        session.close()


@pytest_asyncio.fixture(scope="function")  # pyright: ignore[reportUntypedFunctionDecorator,reportUnknownMemberType]
async def async_session(
    async_scoped_session_maker: async_scoped_session[AsyncSession],
) -> AsyncGenerator[AsyncSession]:
    session = async_scoped_session_maker()
    try:
        yield session
    finally:
        await session.close()


@pytest.fixture(scope="module")
def init_db(
    engine: Engine,
) -> None:
    ItemModel.__table__.create(bind=engine)  # type: ignore - Only for testing
    AsyncItemModel.__table__.create(bind=engine)  # type: ignore - Only for testing
    RecursiveItemModel.__table__.create(bind=engine)  # type: ignore - Only for testing
    SubItemModel.__table__.create(bind=engine)  # type: ignore - Only for testing
    SingularSubItemModel.__table__.create(bind=engine)  # type: ignore - Only for testing
    UserModel.__table__.create(bind=engine)  # type: ignore - Only for testing
    PermissionItemModel.__table__.create(bind=engine)  # type: ignore - Only for testing



from typing import Any

from pydantic import ValidationError
from sqlalchemy.exc import ArgumentError
from sqlalchemy.inspection import inspect

from mixemy.exceptions import MixemyConversionError
from mixemy.models import BaseModel
from mixemy.schemas import BaseSchema
from mixemy.types import BaseModelT, BaseSchemaT


def unpack_schema(
    schema: BaseSchema,
    *,
    exclude_unset: bool = True,
    exclude: set[str] | None = None,
    by_alias: bool = True,
) -> dict[str, Any]:
    return schema.model_dump(
        exclude_unset=exclude_unset, exclude=exclude, by_alias=by_alias
    )


def to_model(
    model: type[BaseModelT],
    schema: BaseSchema,
    *,
    recursive_conversion: bool = False,
    exclude_unset: bool = True,
    exclude: set[str] | None = None,
    by_alias: bool = True,
) -> BaseModelT:
    sub_models: dict[str, list[BaseModel] | BaseModel] = {}
    if recursive_conversion:
        if exclude is None:
            exclude = set()
        relationships = inspect(model).relationships
        for relationship in relationships:
            if hasattr(schema, relationship.key):
                sub_models[relationship.key] = (
                    [
                        to_model(
                            model=relationship.mapper.class_,
                            schema=item,
                            recursive_conversion=recursive_conversion,
                            exclude_unset=exclude_unset,
                            exclude=exclude,
                            by_alias=by_alias,
                        )
                        for item in getattr(schema, relationship.key)
                    ]
                    if relationship.uselist
                    else to_model(
                        model=relationship.mapper.class_,
                        schema=getattr(schema, relationship.key),
                        recursive_conversion=recursive_conversion,
                        exclude_unset=exclude_unset,
                        exclude=exclude,
                        by_alias=by_alias,
                    )
                )

                exclude.add(relationship.key)

    unpacked_schema = unpack_schema(
        schema=schema,
        exclude_unset=exclude_unset,
        exclude=exclude,
        by_alias=by_alias,
    )

    try:
        return model(**unpacked_schema, **sub_models)
    except AttributeError as ex:
        message = (
            f"Error converting {type(schema)} to {model}.\nThis is likely as there is a nested model in {type(schema)} and `recursive_conversion` is false"
            if not recursive_conversion
            else None
        )
        raise MixemyConversionError(
            model=model, schema=type(schema), is_model_to_schema=False, message=message
        ) from ex
    except ArgumentError as ex:
        raise MixemyConversionError(
            model=model, schema=type(schema), is_model_to_schema=False
        ) from ex


def to_schema(model: BaseModel, schema: type[BaseSchemaT]) -> BaseSchemaT:
    try:
        return schema.model_validate(model)
    except ValidationError as ex:
        raise MixemyConversionError(
            model=model, schema=schema, is_model_to_schema=True
        ) from ex



"""This package provides base repository classes for both asynchronous and synchronous operations.

Modules:
    _asyncio: Contains the BaseAsyncRepository class for asynchronous database operations.
    _sync: Contains the BaseSyncRepository class for synchronous database operations.
Classes:
    BaseAsyncRepository: A base class for creating asynchronous repositories.
    PermissionAsyncRepository: A base class for creating asynchronous repositories with permission
    BaseSyncRepository: A base class for creating synchronous repositories.
    PermissionSyncRepository: A base class for creating synchronous repositories with permission
"""

from ._asyncio import BaseAsyncRepository, PermissionAsyncRepository
from ._sync import BaseSyncRepository, PermissionSyncRepository

__all__ = [
    "BaseAsyncRepository",
    "BaseSyncRepository",
    "PermissionAsyncRepository",
    "PermissionSyncRepository",
]



from mixemy.repositories.permission_strategies._base import PermissionStrategy
from mixemy.repositories.permission_strategies._loose_strategy import (
    LoosePermissionStrategy,
)
from mixemy.repositories.permission_strategies._strict_strategy import (
    StrictPermissionStrategy,
)
from mixemy.types import BaseModelT, permission_strategies


class PermissionStrategyFactory:
    """Factory class for creating permission strategies."""

    @staticmethod
    def create_permission_strategy(
        strategy: permission_strategies,
        model: type[BaseModelT],
        user_id_attribute: str,
        user_joined_table: str | None,
    ) -> PermissionStrategy:
        """Create a permission strategy based on the given strategy name."""
        if strategy == "loose":
            return LoosePermissionStrategy(
                model=model,
                user_id_attribute=user_id_attribute,
                user_joined_table=user_joined_table,
            )

        return StrictPermissionStrategy(
            model=model,
            user_id_attribute=user_id_attribute,
            user_joined_table=user_joined_table,
        )



"""This module defines custom exception classes for the Mixemy application.

Classes:
    MixemyError: Base class for all custom exceptions in Mixemy.
    MixemyConversionError: Raised for errors in the conversion between a model and a schema.
    MixemyRepositoryError: Raised for errors that occur within a Mixemy repository.
    MixemyServiceError: Raised for errors that occur within a Mixemy service.
    MixemySetupError: Raised for errors in the setup of a Mixemy component.
    MixemyRepositorySetupError: Raised for errors in the setup of a Mixemy repository.
    MixemyServiceSetupError: Raised for errors in the setup of a Mixemy service.

"""

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from mixemy.models import BaseModel
    from mixemy.repositories import BaseAsyncRepository, BaseSyncRepository
    from mixemy.schemas import BaseSchema
    from mixemy.services import BaseAsyncService, BaseSyncService


class MixemyError(Exception):
    """Custom exception class for Mixemy application.

    Attributes:
        message (str): The error message describing the exception.

    Args:
        message (str): The error message to be displayed.
    """

    def __init__(self, message: str) -> None:
        """Initialize the exception with a given message.

        Args:
            message (str): The error message to be associated with this exception.
        """
        self.message = message
        Exception.__init__(self, message)


class MixemyConversionError(MixemyError):
    """Exception raised for errors in the conversion between a model and a schema.

    Attributes:
        model (BaseModel | type[BaseModel]): The model involved in the conversion.
        schema (BaseSchema | type[BaseSchema]): The schema involved in the conversion.
        is_model_to_schema (bool): Indicates the direction of the conversion.
        message (str | None): Optional error message. If not provided, a default message is generated.

    Args:
        model (BaseModel | type[BaseModel]): The model involved in the conversion.
        schema (BaseSchema | type[BaseSchema]): The schema involved in the conversion.
        is_model_to_schema (bool): Indicates the direction of the conversion.
        message (str | None, optional): Optional error message. Defaults to None.
    """

    def __init__(
        self,
        model: "BaseModel | type[BaseModel]",
        schema: "BaseSchema | type[BaseSchema]",
        is_model_to_schema: bool,
        message: str | None = None,
    ) -> None:
        """Initialize the conversion error between model and schema.

        Args:
            model (BaseModel | type[BaseModel]): The model involved in the conversion.
            schema (BaseSchema | type[BaseSchema]): The schema involved in the conversion.
            is_model_to_schema (bool): Flag indicating the direction of conversion.
                                       True if converting from model to schema, False otherwise.
            message (str | None, optional): Custom error message. Defaults to None.
        """
        if message is None:
            if is_model_to_schema:
                message = f"Error converting {model} to {schema}.\nThis is likely due to a mismatch between the model and schema"
            else:
                message = f"Error converting {schema} to {model}.\nThis is likely due to a mismatch between the schema and model"
        self.model = model
        self.schema = schema
        MixemyError.__init__(self, message)


class MixemyRepositoryError(MixemyError):
    """Exception raised for errors that occur within a Mixemy repository.

    Attributes:
        repository (BaseSyncRepository[Any] | BaseAsyncRepository[Any]): The repository instance where the error occurred.
        message (str): Optional. The error message. If not provided, a default message is generated.

    Args:
        repository (BaseSyncRepository[Any] | BaseAsyncRepository[Any]): The repository instance where the error occurred.
        message (str, optional): The error message. Defaults to None.
    """

    def __init__(
        self,
        repository: "BaseSyncRepository[Any] | BaseAsyncRepository[Any]",
        message: str | None = None,
    ) -> None:
        """Initialize the exception with a repository and an optional message.

        Args:
            repository (BaseSyncRepository[Any] | BaseAsyncRepository[Any]): The repository instance where the error occurred.
            message (str | None, optional): An optional error message. Defaults to None.

        If no message is provided, a default message indicating an error with the repository will be used.
        """
        if message is None:
            message = f"Error with repository {repository}."
        self.repository = repository
        MixemyError.__init__(self, message)


class MixemyRepositoryPermissionError(MixemyRepositoryError):
    """Exception raised when a user does not have the necessary permissions.

    This exception is a subclass of MixemyRepositoryError and is raised when a user
    does not have the necessary permissions to perform an operation.

    Attributes:
        repository (BaseSyncRepository[Any] | BaseAsyncRepository[Any]): The repository instance where the error occurred.
        message (str): The error message.

    Args:
        repository (BaseSyncRepository[Any] | BaseAsyncRepository[Any]): The repository instance where the error occurred.
        message (str): The error message.
    """

    def __init__(
        self,
        repository: "BaseSyncRepository[Any] | BaseAsyncRepository[Any]",
        object_id: Any | None = None,
        user_id: Any | None = None,
        message: str | None = None,
    ) -> None:
        """Initialize the exception with the given repository and message.

        Args:
            repository (BaseSyncRepository[Any] | BaseAsyncRepository[Any]): The repository instance where the error occurred.
            message (str): The error message.
        """
        if message is None:
            message = f"User {user_id} does not have permission to access object {object_id} in repository {repository}."
        self.object_id = object_id
        self.user_id = user_id
        super().__init__(repository=repository, message=message)


class MixemyRepositoryReadOnlyError(MixemyRepositoryError):
    """Exception raised when a repository is in read-only mode.

    This exception is a subclass of MixemyRepositoryError and is raised when an
    operation is attempted on a repository that is in read-only mode.

    Attributes:
        repository (BaseSyncRepository[Any] | BaseAsyncRepository[Any]): The repository instance where the error occurred.
        message (str): The error message.

    Args:
        repository (BaseSyncRepository[Any] | BaseAsyncRepository[Any]): The repository instance where the error occurred.
        message (str): The error message.
    """

    def __init__(
        self,
        repository: "BaseSyncRepository[Any] | BaseAsyncRepository[Any]",
        model: "BaseModel | type[BaseModel] | None" = None,
        operation: str | None = None,
        message: str | None = None,
    ) -> None:
        """Initialize the exception with the given repository and message.

        Args:
            repository (BaseSyncRepository[Any] | BaseAsyncRepository[Any]): The repository instance where the error occurred.
            message (str): The error message.
        """
        if message is None:
            message = f"Repository {repository} is in read-only mode."
            if model is not None:
                message += f" On Model: {model}."
            if operation is not None:
                message += f" Using Operation: {operation}."
        super().__init__(repository=repository, message=message)


class MixemyServiceError(MixemyError):
    """Exception raised for errors that occur within a Mixemy service.

    Attributes:
        service (BaseSyncService[Any, Any] | BaseAsyncService[Any, Any]): The service instance where the error occurred.
        message (str): The error message. If not provided, a default message is generated.

    Args:
        service (BaseSyncService[Any, Any] | BaseAsyncService[Any, Any]): The service instance where the error occurred.
        message (str, optional): The error message. Defaults to None.
    """

    def __init__(
        self,
        service: "BaseSyncService[Any, Any] | BaseAsyncService[Any, Any]",
        message: str | None = None,
    ) -> None:
        """Initialize the exception with the given service and an optional message.

        Args:
            service (BaseSyncService[Any, Any] | BaseAsyncService[Any, Any]):
                The service instance that caused the exception.
            message (str, optional):
                An optional error message. If not provided, a default message
                indicating an error with the service will be used.

        """
        if message is None:
            message = f"Error with service {service}."
        self.service = service
        MixemyError.__init__(self, message)


class MixemySetupError(MixemyError):
    """Exception raised for errors in the setup of a Mixemy component.

    Attributes:
        component (object): The component instance where the error occurred.
        undefined_field (str): The name of the undefined field that caused the error.
        message (str, optional): An optional error message. If not provided, a default message is generated.

    Args:
        component (object): The component instance where the error occurred.
        undefined_field (str): The name of the undefined field that caused the error.
        message (str, optional): An optional error message. If not provided, a default message is generated.
    """

    def __init__(
        self,
        component: object,
        undefined_field: str,
        message: str | None = None,
    ) -> None:
        """Initialize the exception with the component, undefined field, and an optional message.

        Args:
            component (object): The component instance where the undefined field was encountered.
            undefined_field (str): The name of the field that is undefined.
            message (str | None, optional): Custom error message. If not provided, a default message is generated.

        Returns:
            None
        """
        if message is None:
            message = f"{component.__class__.__name__.capitalize()} {component} has undefined field '{undefined_field}'.\nThis probably needs to be defined as a class attribute."
        self.undefined_field = undefined_field
        MixemyError.__init__(self, message)


class MixemyRepositorySetupError(MixemySetupError, MixemyRepositoryError):
    """Exception raised for errors in the setup of a Mixemy repository.

    This exception is a combination of MixemySetupError and MixemyRepositoryError,
    indicating that there was an issue specifically related to the repository setup.

    Attributes:
        repository (BaseSyncRepository[Any] | BaseAsyncRepository[Any]): The repository instance where the error occurred.
        undefined_field (str): The field that was not defined correctly.
        message (str | None): Optional error message providing more details about the error.
    """

    def __init__(
        self,
        repository: "BaseSyncRepository[Any] | BaseAsyncRepository[Any]",
        undefined_field: str,
        message: str | None = None,
    ) -> None:
        """Initialize the exception with the given repository, undefined field, and optional message.

        Args:
            repository (BaseSyncRepository[Any] | BaseAsyncRepository[Any]): The repository instance where the error occurred.
            undefined_field (str): The name of the undefined field that caused the error.
            message (str, optional): An optional error message. Defaults to None.
        """
        MixemySetupError.__init__(
            self,
            component=repository,
            undefined_field=undefined_field,
            message=message,
        )
        MixemyRepositoryError.__init__(
            self, repository=repository, message=self.message
        )


class MixemyServiceSetupError(MixemySetupError, MixemyServiceError):
    """Exception raised for errors in the setup of a Mixemy service.

    This exception is a combination of MixemySetupError and MixemyServiceError,
    indicating that there was an issue with the setup of a service in the Mixemy
    framework.

    Attributes:
        service (BaseSyncService[Any, Any] | BaseAsyncService[Any, Any]): The service
            that encountered the setup error.
        undefined_field (str): The field that was not defined correctly.
        message (str | None): Optional error message providing more details about the
            setup error.

    Args:
        service (BaseSyncService[Any, Any] | BaseAsyncService[Any, Any]): The service
            that encountered the setup error.
        undefined_field (str): The field that was not defined correctly.
        message (str | None, optional): Optional error message providing more details
            about the setup error. Defaults to None.
    """

    def __init__(
        self,
        service: "BaseSyncService[Any, Any] | BaseAsyncService[Any, Any]",
        undefined_field: str,
        message: str | None = None,
    ) -> None:
        """Initialize the exception with the given service, undefined field, and optional message.

        Args:
            service (BaseSyncService[Any, Any] | BaseAsyncService[Any, Any]): The service where the error occurred.
            undefined_field (str): The name of the undefined field that caused the error.
            message (str | None, optional): An optional error message. Defaults to None.
        """
        MixemySetupError.__init__(
            self, component=service, undefined_field=undefined_field, message=message
        )
        MixemyServiceError.__init__(self, service=service, message=self.message)


__all__ = [
    "MixemyConversionError",
    "MixemyError",
    "MixemyRepositoryError",
    "MixemyRepositoryPermissionError",
    "MixemyRepositorySetupError",
    "MixemyServiceError",
    "MixemyServiceSetupError",
    "MixemySetupError",
]



from abc import ABC
from typing import TYPE_CHECKING, Any

from mixemy.exceptions import MixemyServiceSetupError
from mixemy.models import BaseModel
from mixemy.schemas import InputSchema
from mixemy.types import (
    OutputSchemaT,
    PermissionAsyncRepositoryT,
)
from mixemy.utils import to_model, to_schema

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncSession

    from mixemy.repositories import BaseAsyncRepository
    from mixemy.schemas import OutputSchema


class BaseAsyncService[
    RepositoryAsyncT: BaseAsyncRepository[
        Any
    ],  # https://peps.python.org/pep-0695/#explicit-variance
    OutputSchemaT: OutputSchema,
](
    ABC,
):
    """Base class for asynchronous services.

    This class provides a generic implementation for common CRUD operations
    (create, read, update, delete) using asynchronous methods. It is designed
    to work with SQLAlchemy's AsyncSession and generic repository patterns.
    Type Parameters:
        RepositoryAsyncT: The type of the asynchronous repository.
        OutputSchemaT: The type of the schema used for outputting objects.
    Attributes:
        repository_type (type[RepositoryAsyncT]): The type of the repository.
        output_schema_type (type[OutputSchemaT]): The type of the output schema.
        default_recursive_model_conversion (bool): Default value for recursive model conversion.
        default_exclude_unset (bool): Default value for excluding unset fields.
        default_exclude (set[str] | None): Default value for excluded fields.
        default_by_alias (bool): Default value for using field aliases.
    Methods:
        __init__(db_session: AsyncSession) -> None:
            Initializes the service with the given database session.
        create(object_in: CreateSchemaT, *, auto_expunge: bool | None = None, auto_refresh: bool | None = None, auto_commit: bool | None = None) -> OutputSchemaT:
            Asynchronously creates a new object in the database.
        read(id: Any, *, loader_options: tuple[_AbstractLoad] | None = None, execution_options: dict[str, Any] | None = None, auto_expunge: bool | None = None, auto_commit: bool | None = None) -> OutputSchemaT | None:
            Asynchronously reads an object from the database by its ID.
        read_multiple(filters: FilterSchemaT | None = None, *, loader_options: tuple[_AbstractLoad] | None = None, execution_options: dict[str, Any] | None = None, auto_expunge: bool | None = None, auto_commit: bool | None = None) -> list[OutputSchemaT]:
            Asynchronously reads multiple objects from the database based on filters.
        update(id: Any, object_in: UpdateSchemaT, *, loader_options: tuple[_AbstractLoad] | None = None, execution_options: dict[str, Any] | None = None, auto_expunge: bool | None = None, auto_refresh: bool | None = None, auto_commit: bool | None = None) -> OutputSchemaT | None:
            Asynchronously updates an object in the database by its ID.
        delete(id: Any, *, loader_options: tuple[_AbstractLoad] | None = None, execution_options: dict[str, Any] | None = None, auto_expunge: bool | None = None, auto_commit: bool | None = None) -> None:
            Asynchronously deletes an object from the database by its ID.
        _to_model(schema: CreateSchemaT | UpdateSchemaT) -> BaseModelT:
            Converts a schema to a model instance.
        _to_schema(model: BaseModelT) -> OutputSchemaT:
            Converts a model instance to a schema.
        _verify_init() -> None:
            Verifies that the required attributes are set during initialization.
    """

    repository_type: type[RepositoryAsyncT]
    output_schema_type: type[OutputSchemaT]

    default_recursive_model_conversion: bool = False
    default_exclude_unset: bool = True
    default_exclude: set[str] | None = None
    default_by_alias: bool = True

    def __init__(
        self,
        db_session: "AsyncSession",
        *,
        recursive_model_conversion: bool | None = None,
        exclude_unset: bool | None = None,
        exclude: set[str] | None = None,
        by_alias: bool | None = None,
    ) -> None:
        self._verify_init()
        self.repository = self.repository_type()
        self.model = self.repository.model
        self.output_schema = self.output_schema_type
        self.db_session = db_session
        self.recursive_model_conversion = (
            recursive_model_conversion
            if recursive_model_conversion is not None
            else self.default_recursive_model_conversion
        )
        self.exclude_unset = (
            exclude_unset if exclude_unset is not None else self.default_exclude_unset
        )
        self.exclude = exclude if exclude is not None else self.default_exclude
        self.by_alias = by_alias if by_alias is not None else self.default_by_alias

    async def create(
        self,
        object_in: InputSchema,
        *,
        recursive_model_conversion: bool | None = None,
        exclude_unset: bool | None = None,
        exclude: set[str] | None = None,
        by_alias: bool | None = None,
        **kwargs: Any,
    ) -> OutputSchemaT:
        return self.to_schema(
            model=await self.repository.create(
                db_session=self.db_session,
                db_object=self.to_model(
                    schema=object_in,
                    recursive_model_conversion=recursive_model_conversion,
                    by_alias=by_alias,
                    exclude_unset=exclude_unset,
                    exclude=exclude,
                ),
                **kwargs,
            ),
        )

    async def read(
        self,
        id: Any,
        **kwargs: Any,
    ) -> OutputSchemaT | None:
        return (
            self.to_schema(model=model)
            if (
                model := await self.repository.read(
                    db_session=self.db_session,
                    id=id,
                    **kwargs,
                )
            )
            else None
        )

    async def read_multiple(
        self,
        filters: InputSchema | None = None,
        **kwargs: Any,
    ) -> list[OutputSchemaT]:
        return [
            self.to_schema(model=model)
            for model in await self.repository.read_multiple(
                db_session=self.db_session,
                filters=filters,
                **kwargs,
            )
        ]

    async def update(
        self,
        id: Any,
        object_in: InputSchema,
        **kwargs: Any,
    ) -> OutputSchemaT | None:
        return (
            self.to_schema(model)
            if (
                model := await self.repository.update(
                    db_session=self.db_session,
                    id=id,
                    object_in=object_in,
                    **kwargs,
                )
            )
            else None
        )

    async def delete(
        self,
        id: Any,
        **kwargs: Any,
    ) -> None:
        await self.repository.delete(
            db_session=self.db_session,
            id=id,
            **kwargs,
        )

    def to_model(
        self,
        schema: InputSchema,
        *,
        recursive_model_conversion: bool | None = None,
        exclude_unset: bool | None = None,
        exclude: set[str] | None = None,
        by_alias: bool | None = None,
    ) -> BaseModel:
        current_recursive_model_conversion = (
            recursive_model_conversion
            if recursive_model_conversion is not None
            else self.recursive_model_conversion
        )
        current_exclude_unset = (
            exclude_unset if exclude_unset is not None else self.exclude_unset
        )
        current_exclude = exclude if exclude is not None else self.exclude
        current_by_alias = by_alias if by_alias is not None else self.by_alias
        return to_model(
            schema=schema,
            model=self.model,
            recursive_conversion=current_recursive_model_conversion,
            exclude_unset=current_exclude_unset,
            exclude=current_exclude,
            by_alias=current_by_alias,
        )

    def to_schema(self, model: BaseModel) -> OutputSchemaT:
        return to_schema(model=model, schema=self.output_schema)

    def _verify_init(self) -> None:
        for field in ["output_schema_type", "repository_type"]:
            if not hasattr(self, field):
                raise MixemyServiceSetupError(service=self, undefined_field=field)


class PermissionAsyncService(
    BaseAsyncService[PermissionAsyncRepositoryT, OutputSchemaT]
):
    """Service for performing asynchronous operations on Permission objects.

    This class provides methods for creating, reading, updating, and deleting
    Permission objects using asynchronous methods. It is designed to work with
    SQLAlchemy's AsyncSession and the PermissionAsyncRepository.
    Methods:
        __init__(db_session: AsyncSession) -> None:
            Initializes the service with the given database session.
    """

    repository_type: type[PermissionAsyncRepositoryT]

    def __init__(
        self,
        db_session: "AsyncSession",
        *,
        recursive_model_conversion: bool | None = None,
        exclude_unset: bool | None = None,
        exclude: set[str] | None = None,
        by_alias: bool | None = None,
    ) -> None:
        super().__init__(
            db_session=db_session,
            recursive_model_conversion=recursive_model_conversion,
            exclude_unset=exclude_unset,
            exclude=exclude,
            by_alias=by_alias,
        )

    async def read_with_permission(
        self,
        id: Any,
        user_id: Any,
        **kwargs: Any,
    ) -> OutputSchemaT | None:
        return (
            self.to_schema(model=model)
            if (
                model := await self.repository.read_with_permission(
                    db_session=self.db_session,
                    id=id,
                    user_id=user_id,
                    **kwargs,
                )
            )
            else None
        )

    async def read_multiple_with_permission(
        self,
        user_id: Any,
        filters: InputSchema | None = None,
        **kwargs: Any,
    ) -> list[OutputSchemaT]:
        return [
            self.to_schema(model=model)
            for model in await self.repository.read_multiple_with_permission(
                db_session=self.db_session,
                user_id=user_id,
                filters=filters,
                **kwargs,
            )
        ]

    async def update_with_permission(
        self,
        id: Any,
        object_in: InputSchema,
        user_id: Any,
        **kwargs: Any,
    ) -> OutputSchemaT | None:
        return (
            self.to_schema(model)
            if (
                model := await self.repository.update_with_permission(
                    db_session=self.db_session,
                    id=id,
                    user_id=user_id,
                    object_in=object_in,
                    **kwargs,
                )
            )
            else None
        )

    async def delete_with_permission(
        self,
        id: Any,
        user_id: Any,
        **kwargs: Any,
    ) -> None:
        await self.repository.delete_with_permission(
            db_session=self.db_session,
            id=id,
            user_id=user_id,
            **kwargs,
        )



"""This module provides utility functions for the mixemy package.

Functions:
    generate_uuid: Generates a unique identifier (UUID).
    to_model: Converts data to a model instance.
    to_schema: Converts data to a schema instance.
    unpack_schema: Unpacks a schema into its components.
    unpack_sequence: Unpacks a value or a sequence into a sequence.
"""

from ._convertors import to_model, to_schema, unpack_schema
from ._utils import pack_sequence
from ._uuids import generate_uuid

__all__ = ["generate_uuid", "pack_sequence", "to_model", "to_schema", "unpack_schema"]



from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import DeclarativeBase


class BaseModel(DeclarativeBase):
    """
    BaseModel is an abstract base class for SQLAlchemy declarative models.

    Methods:
        __tablename__(cls) -> str:
            Generates a table name for the model by converting the class name from CamelCase to snake_case.
            Removes the suffix "_model" from the generated table name.
        __repr__(self) -> str:
            Returns a string representation of the model instance, including the class name and its attributes.
    """

    __abstract__ = True

    @declared_attr  # pyright: ignore[reportArgumentType]
    @classmethod
    def __tablename__(cls) -> str:
        return "".join(
            (letter if letter.islower() else "_" + letter.lower())
            for letter in (cls.__name__[0].lower() + cls.__name__[1:])
        ).removesuffix("_model")

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.__dict__}>"



from abc import ABC, abstractmethod
from typing import Any

from sqlalchemy import Select

from mixemy.types import BaseModelT, SelectT


class PermissionStrategy(ABC):
    def __init__(
        self,
        model: type[BaseModelT],
        user_id_attribute: str,
        user_joined_table: str | None,
    ) -> None:
        self.model = model
        self.user_id_attribute = user_id_attribute
        self.user_joined_table = user_joined_table

    @abstractmethod
    def add_permission_filter(
        self,
        statement: Select[SelectT],
        user_id: Any,
    ) -> Select[SelectT]: ...

    @abstractmethod
    def check_permission_on_db_oject(
        self,
        db_object: BaseModelT | None,  # type: ignore no
        user_id: Any,
    ) -> tuple[bool, Any]: ...



"""This module initializes the services package by importing and exposing the BaseAsyncService and BaseSyncService classes.

Classes:
    BaseAsyncService: A base class for asynchronous services.
    BaseSyncService: A base class for synchronous services.
__all__:
    A list of public objects of this module, as interpreted by `import *`.
"""

from ._asyncio import BaseAsyncService, PermissionAsyncService
from ._sync import BaseSyncService, PermissionSyncService

__all__ = [
    "BaseAsyncService",
    "BaseSyncService",
    "PermissionAsyncService",
    "PermissionSyncService",
]



"""mixemy package.

This package provides a set of base models, repositories, and services for building applications.
It includes both asynchronous and synchronous versions of repositories and services.
Modules:
    models: Contains the base models used in the application.
    repositories: Contains the base repository classes for database operations.
    schemas: Contains the schemas used for data validation and serialization.
    utils: Contains utility functions and classes.
Classes:
    BaseModel: The base model class for all database models.
    IdAuditModel: A base model class that includes ID and audit fields.
    BaseAsyncRepository: The base class for asynchronous repository operations.
    BaseSyncRepository: The base class for synchronous repository operations.
    BaseAsyncService: The base class for asynchronous service operations.
    BaseSyncService: The base class for synchronous service operations.
Attributes:
    __all__: A list of public objects of this package.
    __version__: The version of the mixemy package.
"""

from importlib import metadata

from . import exceptions, models, repositories, schemas, utils
from .models import BaseModel, IdAuditModel
from .repositories import (
    BaseAsyncRepository,
    BaseSyncRepository,
    PermissionAsyncRepository,
    PermissionSyncRepository,
)
from .services import (
    BaseAsyncService,
    BaseSyncService,
    PermissionAsyncService,
    PermissionSyncService,
)

__all__ = [
    "BaseAsyncRepository",
    "BaseAsyncService",
    "BaseModel",
    "BaseSyncRepository",
    "BaseSyncService",
    "IdAuditModel",
    "PermissionAsyncRepository",
    "PermissionAsyncService",
    "PermissionSyncRepository",
    "PermissionSyncService",
    "exceptions",
    "models",
    "repositories",
    "schemas",
    "utils",
]
__version__ = metadata.version("mixemy")



from collections.abc import Sequence


def pack_sequence[T](value: T | Sequence[T]) -> Sequence[T]:
    """Pack a value or a sequence into a sequence."""
    if isinstance(value, Sequence):
        return value  # pyright: ignore[reportUnknownVariableType] WTF

    return [value]



from typing import TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
    from typing import Any

    from mixemy.repositories import (
        BaseAsyncRepository,
        BaseSyncRepository,
        PermissionAsyncRepository,
        PermissionSyncRepository,
    )

RepositorySyncT = TypeVar("RepositorySyncT", bound="BaseSyncRepository[Any]")
RepositoryAsyncT = TypeVar("RepositoryAsyncT", bound="BaseAsyncRepository[Any]")
PermissionSyncRepositoryT = TypeVar(
    "PermissionSyncRepositoryT", bound="PermissionSyncRepository[Any]"
)
PermissionAsyncRepositoryT = TypeVar(
    "PermissionAsyncRepositoryT", bound="PermissionAsyncRepository[Any]"
)



"""This module initializes and exports the models used in the mixemy application.

Modules:
    _audit: Contains the AuditModel class.
    _base: Contains the BaseModel class.
    _id: Contains the IdModel class.
    _id_audit: Contains the IdAuditModel class.
Exports:
    AuditModel: A model for auditing purposes.
    BaseModel: The base model for other models to inherit from.
    IdModel: A model that includes an ID field.
    IdAuditModel: A model that combines ID and auditing functionalities.
"""

from ._audit import AuditModel
from ._base import BaseModel
from ._id import IdModel
from ._id_audit import IdAuditModel

__all__ = ["AuditModel", "BaseModel", "IdAuditModel", "IdModel"]



"""This module provides pagination-related classes and constants for the Mixemy application.

Classes:
    AuditPaginationFilter: A filter class for auditing pagination.
    PaginationFilter: A base class for pagination filters.
    OrderBy: An enumeration for specifying the field to order by.
    OrderDirection: An enumeration for specifying the direction of ordering.
Constants:
    PaginationFields: A set of fields used for pagination, including 'limit', 'offset', 'order_by', and 'order_direction'.
__all__:
    A list of public objects of this module, as interpreted by 'import *'.
"""

from ._audit import AuditPaginationFilter
from ._base import PaginationFilter
from ._order_enums import OrderBy, OrderDirection

PaginationFields = {"limit", "offset", "order_by", "order_direction"}

__all__ = [
    "AuditPaginationFilter",
    "OrderBy",
    "OrderDirection",
    "PaginationFields",
    "PaginationFilter",
]



from mixemy.models._audit import AuditModel
from mixemy.models._id import IdModel


class IdAuditModel(IdModel, AuditModel):
    """IdAuditModel is an abstract base class that combines the functionality of IdModel and AuditModel.

    Attributes:
        id (int): The unique identifier for the model instance.
        created_at (datetime): The timestamp when the model instance was created.
        updated_at (datetime): The timestamp when the model instance was last updated.
    """

    __abstract__ = True

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} created_at={self.created_at} updated_at={self.updated_at}>"



from sqlalchemy import UUID as SQLAUUID
from sqlalchemy.orm import Mapped, mapped_column
from uuid_utils.compat import UUID

from mixemy.models._base import BaseModel
from mixemy.utils import generate_uuid


class IdModel(BaseModel):
    """IdModel is an abstract base model that provides a UUID primary key for derived models.

    Attributes:
        id (UUID): The primary key for the model, automatically generated using the `generate_uuid` function.
    """

    __abstract__ = True

    id: Mapped[UUID] = mapped_column(SQLAUUID, primary_key=True, default=generate_uuid)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id}>"



from typing import TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
    from mixemy.schemas import BaseSchema, InputSchema, OutputSchema

BaseSchemaT = TypeVar("BaseSchemaT", bound="BaseSchema")
CreateSchemaT = TypeVar("CreateSchemaT", bound="InputSchema")
UpdateSchemaT = TypeVar("UpdateSchemaT", bound="InputSchema")
OutputSchemaT = TypeVar("OutputSchemaT", bound="OutputSchema")

FilterSchemaT = TypeVar("FilterSchemaT", bound="InputSchema")



from typing import TYPE_CHECKING

import pytest
from sqlalchemy.ext.asyncio import AsyncSession

if TYPE_CHECKING:
    from mixemy.models import IdAuditModel


@pytest.mark.database
@pytest.mark.integration
@pytest.mark.asyncio
async def test_main(
    async_session: AsyncSession, async_item_model: "type[IdAuditModel]", init_db: None
) -> None:
    """Test the main functionality of the asynchronous item service.

    This test function performs the following operations:
    1. Creates two items with values "test_one" and "test_two".
    2. Asserts that the created items have the correct values.
    3. Reads the created items from the database and asserts their existence and values.
    4. Updates the first item with the value of the second item and asserts the update.
    5. Deletes the first item and asserts its deletion.
    6. Reads the second item and asserts its existence and value.
    7. Deletes the second item and asserts its deletion.

    Args:
        async_session (AsyncSession): The asynchronous database session.
        async_item_model (type[IdAuditModel]): The asynchronous item model type.
        init_db (None): A placeholder for database initialization.
    """
    from mixemy import repositories, schemas, services

    AsyncItemModel = async_item_model

    class ItemInput(schemas.InputSchema):
        value: str

    class ItemOutput(schemas.IdAuditOutputSchema):
        value: str

    class ItemRepository(repositories.BaseAsyncRepository[AsyncItemModel]):
        model = AsyncItemModel

    class ItemService(services.BaseAsyncService[ItemRepository, ItemOutput]):
        repository_type = ItemRepository
        output_schema_type = ItemOutput

    item_service = ItemService(db_session=async_session)

    test_one = ItemInput(value="test_one")
    test_two = ItemInput(value="test_two")

    item_one = await item_service.create(object_in=test_one)
    item_two = await item_service.create(object_in=test_two)

    assert item_one.value == "test_one"
    assert item_two.value == "test_two"

    item_one = await item_service.read(id=item_one.id)
    item_two = await item_service.read(id=item_two.id)

    assert item_one is not None
    assert item_two is not None
    assert item_one.value == "test_one"
    assert item_two.value == "test_two"

    item_one = await item_service.update(id=item_one.id, object_in=test_two)

    assert item_one is not None
    assert item_one.value == "test_two"

    await item_service.delete(id=item_one.id)

    item_one = await item_service.read(id=item_one.id)
    item_two = await item_service.read(id=item_two.id)

    assert item_one is None
    assert item_two is not None
    assert item_two.value == "test_two"

    await item_service.delete(id=item_two.id)

    item_two = await item_service.read(id=item_two.id)

    assert item_two is None



"""This package initializes and exports various schema modules used in the application.

Modules:
    paginations: Handles pagination schemas.
    serializers: Handles serialization schemas.
    AuditOutputSchema: Schema for audit output.
    BaseSchema: Base schema class.
    IdAuditOutputSchema: Schema for ID audit output.
    IdOutputSchema: Schema for ID output.
    InputSchema: Schema for input data.
    OutputSchema: Schema for output data.
Exports:
    AuditOutputSchema
    BaseSchema
    IdAuditOutputSchema
    IdOutputSchema
    InputSchema
    OutputSchema
    paginations
    serializers
"""

from . import paginations
from ._audit_output import AuditOutputSchema
from ._base import BaseSchema
from ._id_audit_output import IdAuditOutputSchema
from ._id_output import IdOutputSchema
from ._input import InputSchema
from ._output import OutputSchema

__all__ = [
    "AuditOutputSchema",
    "BaseSchema",
    "IdAuditOutputSchema",
    "IdOutputSchema",
    "InputSchema",
    "OutputSchema",
    "paginations",
]



from abc import ABC
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Literal, overload

from sqlalchemy import (
    CursorResult,
    Delete,
    Result,
    Row,
    ScalarResult,
    Select,
    Update,
    func,
    select,
)
from sqlalchemy.orm.strategy_options import (
    _AbstractLoad,  # pyright: ignore[reportPrivateUsage]
)
from sqlalchemy.util import EMPTY_DICT

from mixemy.exceptions import (
    MixemyRepositoryPermissionError,
    MixemyRepositoryReadOnlyError,
    MixemyRepositorySetupError,
)
from mixemy.repositories.permission_strategies import PermissionStrategyFactory
from mixemy.schemas import InputSchema
from mixemy.schemas.paginations import PaginationFields, PaginationFilter
from mixemy.types import BaseModelT, SelectT, permission_strategies
from mixemy.utils import pack_sequence, unpack_schema

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from mixemy.models import BaseModel


class BaseSyncRepository[BaseModelT: BaseModel](ABC):
    """Base synchronous repository class providing CRUD operations for a given SQLAlchemy model.

    Attributes:
        model (type[BaseModelT]): The SQLAlchemy model type.
        default_loader_options (tuple[_AbstractLoad] | None): Default loader options for SQLAlchemy queries.
        default_execution_options (dict[str, Any] | None): Default execution options for SQLAlchemy queries.
        default_auto_expunge (bool): Whether to automatically expunge objects from the session after operations. Defaults to False.
        default_auto_refresh (bool): Whether to automatically refresh objects after operations. Defaults to True.
        default_auto_commit (bool): Whether to automatically commit the session after operations. Defaults to False.
        default_is_read_only (bool): Whether the repository is in read-only mode. Defaults to False.
    Methods:
        __init__(self, *, loader_options=None, execution_options=None, auto_expunge=False, auto_refresh=True, auto_commit=False):
            Initializes the repository with optional custom settings.
        create(self, db_session, db_object, *, auto_commit=None, auto_expunge=None, auto_refresh=None):
            Synchronously creates a new object in the database.
        read(self, db_session, id, *, loader_options=None, execution_options=None, auto_expunge=None, auto_commit=False, with_for_update=False):
            Synchronously reads an object from the database by its ID.
        read_multiple(self, db_session, filters=None, *, loader_options=None, execution_options=None, auto_commit=False, auto_expunge=None, with_for_update=False):
            Synchronously reads multiple objects from the database based on filters.
        update(self, db_session, id, object_in, *, loader_options=None, execution_options=None, auto_commit=None, auto_expunge=None, auto_refresh=None, with_for_update=True):
            Synchronously updates an object in the database by its ID.
        update_db_object(self, db_session, db_object, object_in, *, auto_commit=None, auto_expunge=None, auto_refresh=None):
            Synchronously updates an existing database object.
        delete(self, db_session, id, *, loader_options=None, execution_options=None, auto_commit=None, auto_expunge=None, with_for_update=False):
            Synchronously deletes an object from the database by its ID.
        delete_db_object(self, db_session, db_object, *, auto_commit=None, auto_expunge=None):
            Synchronously deletes an existing database object.
        count(self, db_session, filters=None, *, loader_options=None, execution_options=None, auto_commit=False):
            Synchronously counts the number of objects in the database based on filters.
        _maybe_commit_or_flush_or_refresh_or_expunge(self, db_session, db_object, *, auto_commit, auto_expunge, auto_refresh):
            Helper method to commit, flush, refresh, or expunge objects from the session based on settings.
        _add(self, db_session, db_object, *, auto_commit, auto_expunge, auto_refresh):
            Helper method to add an object to the session and handle post-add operations.
        _delete(self, db_session, db_object, *, auto_commit, auto_expunge):
            Helper method to delete an object from the session and handle post-delete operations.
        _get(self, db_session, id, *, loader_options, execution_options, auto_expunge, with_for_update, auto_commit=False):
            Helper method to get an object from the database by its ID.
        execute_returning_all(self, db_session, statement, *, loader_options, execution_options, auto_commit, auto_expunge, auto_refresh, with_for_update=False):
            Helper method to execute a statement and return all results.
        execute_returning_one(self, db_session, statement, *, loader_options, execution_options, auto_commit, auto_expunge, auto_refresh, with_for_update=False):
            Helper method to execute a statement and return a single result.
        def add_pagination(self, statement, filters):
            Helper method to add pagination to a SQLAlchemy statement.
        def add_filters(self, statement, filters):
            Helper method to add filters to a SQLAlchemy statement.
        def _prepare_statement(self, statement, *, loader_options, execution_options, with_for_update):
            Helper method to prepare a SQLAlchemy statement with options and execution settings.
        def _update_db_object(db_object, object_in):
            Helper method to update a database object with new values.
        def handle_read_only(self, is_read_only=None):
            Helper method to handle read-only mode for the repository.
        def _verify_init(self):
            Helper method to verify that required attributes are set during initialization.
    """

    model: type[BaseModelT]

    default_loader_options: tuple[_AbstractLoad] | None = None
    default_execution_options: dict[str, Any] | None = None
    default_auto_expunge: bool = False
    default_auto_refresh: bool = True
    default_auto_commit: bool = False
    default_is_read_only: bool = False

    def __init__(
        self,
        *,
        loader_options: tuple[_AbstractLoad] | None = None,
        execution_options: dict[str, Any] | None = None,
        auto_expunge: bool | None = None,
        auto_refresh: bool | None = None,
        auto_commit: bool | None = None,
        is_read_only: bool | None = None,
    ) -> None:
        self.loader_options = (
            loader_options
            if loader_options is not None
            else self.default_loader_options
        )
        self.execution_options = (
            execution_options
            if execution_options is not None
            else self.default_execution_options
        )
        self.auto_expunge = (
            auto_expunge if auto_expunge is not None else self.default_auto_expunge
        )
        self.auto_refresh = (
            auto_refresh if auto_refresh is not None else self.default_auto_refresh
        )
        self.auto_commit = (
            auto_commit if auto_commit is not None else self.default_auto_commit
        )
        self.is_read_only = (
            is_read_only if is_read_only is not None else self.default_is_read_only
        )
        self._verify_init()

    def create(
        self,
        db_session: "Session",
        db_object: BaseModelT,
        *,
        auto_commit: bool | None = None,
        auto_expunge: bool | None = None,
        auto_refresh: bool | None = None,
        is_read_only: bool | None = None,
    ) -> BaseModelT:
        self.handle_read_only(is_read_only=is_read_only)
        return self._add(
            db_session=db_session,
            db_object=db_object,
            auto_commit=auto_commit,
            auto_expunge=auto_expunge,
            auto_refresh=auto_refresh,
        )

    def read(
        self,
        db_session: "Session",
        id: Any,
        *,
        loader_options: tuple[_AbstractLoad] | None = None,
        execution_options: dict[str, Any] | None = None,
        auto_expunge: bool | None = None,
        auto_commit: bool | None = False,
        with_for_update: bool = False,
    ) -> BaseModelT | None:
        return self._get(
            db_session=db_session,
            id=id,
            loader_options=loader_options,
            execution_options=execution_options,
            auto_expunge=auto_expunge,
            auto_commit=auto_commit,
            with_for_update=with_for_update,
        )

    def read_multiple(
        self,
        db_session: "Session",
        filters: InputSchema | None = None,
        *,
        loader_options: tuple[_AbstractLoad] | None = None,
        execution_options: dict[str, Any] | None = None,
        auto_commit: bool | None = False,
        auto_expunge: bool | None = None,
        with_for_update: bool = False,
    ) -> Sequence[BaseModelT]:
        statement = select(self.model)
        statement = self.add_filters(statement=statement, filters=filters)
        statement = self.add_pagination(statement=statement, filters=filters)
        return self.execute_returning_all(
            db_session=db_session,
            statement=statement,
            loader_options=loader_options,
            execution_options=execution_options,
            auto_commit=auto_commit,
            auto_expunge=auto_expunge,
            auto_refresh=False,
            with_for_update=with_for_update,
        )

    def update(
        self,
        db_session: "Session",
        id: Any,
        object_in: InputSchema,
        *,
        loader_options: tuple[_AbstractLoad] | None = None,
        execution_options: dict[str, Any] | None = None,
        auto_commit: bool | None = None,
        auto_expunge: bool | None = None,
        auto_refresh: bool | None = None,
        with_for_update: bool = True,
        is_read_only: bool | None = None,
    ) -> BaseModelT | None:
        db_object = self._get(
            db_session=db_session,
            id=id,
            loader_options=loader_options,
            execution_options=execution_options,
            auto_expunge=False,
            with_for_update=with_for_update,
            auto_commit=False,
        )

        return self.update_db_object(
            db_session=db_session,
            db_object=db_object,
            object_in=object_in,
            auto_commit=auto_commit,
            auto_expunge=auto_expunge,
            auto_refresh=auto_refresh,
            is_read_only=is_read_only,
        )

    def update_db_object(
        self,
        db_session: "Session",
        db_object: BaseModelT | None,
        object_in: InputSchema,
        *,
        auto_commit: bool | None = None,
        auto_expunge: bool | None = None,
        auto_refresh: bool | None = None,
        is_read_only: bool | None = None,
    ) -> BaseModelT | None:
        self.handle_read_only(is_read_only=is_read_only)
        if db_object is not None:
            self._update_db_object(db_object=db_object, object_in=object_in)
            return self._add(
                db_session=db_session,
                db_object=db_object,
                auto_commit=auto_commit,
                auto_expunge=auto_expunge,
                auto_refresh=auto_refresh,
            )

        self._maybe_commit_or_flush_or_refresh_or_expunge(
            db_session=db_session,
            db_object=db_object,
            auto_commit=auto_commit,
            auto_expunge=auto_expunge,
            auto_refresh=auto_refresh,
        )

        return None

    def delete(
        self,
        db_session: "Session",
        id: Any,
        *,
        loader_options: tuple[_AbstractLoad] | None = None,
        execution_options: dict[str, Any] | None = None,
        auto_commit: bool | None = None,
        auto_expunge: bool | None = None,
        with_for_update: bool = False,
        is_read_only: bool | None = None,
    ) -> None:
        db_object = self._get(
            db_session=db_session,
            id=id,
            loader_options=loader_options,
            execution_options=execution_options,
            auto_expunge=False,
            auto_commit=False,
            with_for_update=with_for_update,
        )
        self.delete_db_object(
            db_session=db_session,
            db_object=db_object,
            auto_commit=auto_commit,
            auto_expunge=auto_expunge,
            is_read_only=is_read_only,
        )

    def delete_db_object(
        self,
        db_session: "Session",
        db_object: BaseModelT | None,
        *,
        auto_commit: bool | None = None,
        auto_expunge: bool | None = None,
        is_read_only: bool | None = None,
    ) -> None:
        self.handle_read_only(is_read_only=is_read_only)
        if db_object is not None:
            self._delete(
                db_session=db_session,
                db_object=db_object,
                auto_commit=auto_commit,
                auto_expunge=auto_expunge,
            )
        else:
            self._maybe_commit_or_flush_or_refresh_or_expunge(
                db_session=db_session,
                db_object=db_object,
                auto_commit=auto_commit,
                auto_expunge=auto_expunge,
                auto_refresh=False,
            )

    def count(
        self,
        db_session: "Session",
        filters: InputSchema | None = None,
        *,
        loader_options: tuple[_AbstractLoad] | None = None,
        execution_options: dict[str, Any] | None = None,
        auto_commit: bool | None = False,
    ) -> int:
        statement = select(func.count()).select_from(self.model)
        statement = self.add_filters(statement=statement, filters=filters)
        return self.execute_returning_one(
            db_session=db_session,
            statement=statement,
            loader_options=loader_options,
            execution_options=execution_options,
            auto_commit=auto_commit,
            auto_expunge=False,
            auto_refresh=False,
            with_for_update=False,
        )

    def _maybe_commit_or_flush_or_refresh_or_expunge(
        self,
        db_session: "Session",
        db_object: Sequence[object] | object | None,
        *,
        auto_commit: bool | None = None,
        auto_expunge: bool | None = None,
        auto_refresh: bool | None = None,
        refresh_columns: list[str] | None = None,
    ) -> None:
        if auto_commit is True or (auto_commit is None and self.auto_commit):
            db_session.commit()
        else:
            db_session.flush()

        if db_object is not None:
            instances = pack_sequence(db_object)
            if auto_refresh is True or (auto_refresh is None and self.auto_refresh):
                for instance in instances:
                    db_session.refresh(instance, attribute_names=refresh_columns)
            if auto_expunge is True or (auto_expunge is None and self.auto_expunge):
                for instance in instances:
                    db_session.expunge(instance)

    def _add(
        self,
        db_session: "Session",
        db_object: BaseModelT,
        *,
        auto_commit: bool | None = None,
        auto_expunge: bool | None = None,
        auto_refresh: bool | None = None,
    ) -> BaseModelT:
        db_session.add(db_object)
        self._maybe_commit_or_flush_or_refresh_or_expunge(
            db_session=db_session,
            db_object=db_object,
            auto_commit=auto_commit,
            auto_expunge=auto_expunge,
            auto_refresh=auto_refresh,
        )

        return db_object

    def _delete(
        self,
        db_session: "Session",
        db_object: BaseModelT,
        *,
        auto_commit: bool | None = None,
        auto_expunge: bool | None = None,
    ) -> None:
        db_session.delete(db_object)
        self._maybe_commit_or_flush_or_refresh_or_expunge(
            db_session=db_session,
            db_object=db_object,
            auto_commit=auto_commit,
            auto_expunge=auto_expunge,
            auto_refresh=False,
        )

    def _get(
        self,
        db_session: "Session",
        id: Any,
        *,
        loader_options: tuple[_AbstractLoad] | None = None,
        execution_options: dict[str, Any] | None = None,
        auto_expunge: bool | None = None,
        with_for_update: bool = False,
        auto_commit: bool | None = False,
    ) -> BaseModelT | None:
        current_loader_options = (
            loader_options if loader_options is not None else self.loader_options
        )
        current_execution_options = (
            execution_options
            if execution_options is not None
            else self.execution_options
        )
        db_object = db_session.get(
            self.model,
            id,
            options=current_loader_options,
            execution_options=current_execution_options or EMPTY_DICT,
            with_for_update=with_for_update,
        )
        self._maybe_commit_or_flush_or_refresh_or_expunge(
            db_session=db_session,
            db_object=db_object,
            auto_commit=auto_commit,
            auto_expunge=auto_expunge,
            auto_refresh=False,
        )

        return db_object

    @overload
    def _execute(
        self,
        db_session: "Session",
        statement: Select[SelectT],
        *,
        loader_options: tuple[_AbstractLoad] | None,
        execution_options: dict[str, Any] | None,
        with_for_update: bool,
        is_scalar: Literal[False],
        auto_commit: bool | None,
    ) -> Result[SelectT]: ...

    @overload
    def _execute(
        self,
        db_session: "Session",
        statement: Select[SelectT] | Delete | Update,
        *,
        loader_options: tuple[_AbstractLoad] | None,
        execution_options: dict[str, Any] | None,
        with_for_update: bool,
        is_scalar: Literal[True] = True,
        auto_commit: bool | None,
    ) -> ScalarResult[Any]: ...

    @overload
    def _execute(
        self,
        db_session: "Session",
        statement: Delete | Update,
        *,
        loader_options: tuple[_AbstractLoad] | None,
        execution_options: dict[str, Any] | None,
        with_for_update: bool,
        is_scalar: Literal[False],
        auto_commit: bool | None,
    ) -> CursorResult[Any]: ...

    def _execute(
        self,
        db_session: "Session",
        statement: Select[SelectT] | Delete | Update,
        *,
        loader_options: tuple[_AbstractLoad] | None = None,
        execution_options: dict[str, Any] | None = None,
        with_for_update: bool = False,
        is_scalar: bool = True,
        auto_commit: bool | None = False,
    ) -> Result[SelectT] | ScalarResult[Any] | CursorResult[Any]:
        statement = self._prepare_statement(
            statement=statement,
            loader_options=loader_options,
            execution_options=execution_options,
            with_for_update=with_for_update,
        )
        res = db_session.execute(statement)

        self._maybe_commit_or_flush_or_refresh_or_expunge(
            db_session=db_session,
            db_object=None,
            auto_commit=auto_commit,
            auto_expunge=False,
            auto_refresh=False,
        )

        return res if not is_scalar else res.scalars()

    @overload
    def execute_returning_all(
        self,
        db_session: "Session",
        statement: Select[SelectT],
        *,
        loader_options: tuple[_AbstractLoad] | None,
        execution_options: dict[str, Any] | None,
        auto_commit: bool | None,
        auto_expunge: bool | None,
        auto_refresh: bool | None,
        with_for_update: bool,
        is_scalar: Literal[False],
    ) -> Sequence[SelectT]: ...

    @overload
    def execute_returning_all(
        self,
        db_session: "Session",
        statement: Select[SelectT] | Update | Delete,
        *,
        loader_options: tuple[_AbstractLoad] | None,
        execution_options: dict[str, Any] | None,
        auto_commit: bool | None,
        auto_expunge: bool | None,
        auto_refresh: bool | None,
        with_for_update: bool,
        is_scalar: Literal[True] = True,
    ) -> Sequence[Any]: ...

    @overload
    def execute_returning_all(
        self,
        db_session: "Session",
        statement: Update | Delete,
        *,
        loader_options: tuple[_AbstractLoad] | None,
        execution_options: dict[str, Any] | None,
        auto_commit: bool | None,
        auto_expunge: bool | None,
        auto_refresh: bool | None,
        with_for_update: bool,
        is_scalar: Literal[False],
    ) -> Sequence[Row[Any]]: ...

    def execute_returning_all(
        self,
        db_session: "Session",
        statement: Select[SelectT] | Update | Delete,
        *,
        loader_options: tuple[_AbstractLoad] | None = None,
        execution_options: dict[str, Any] | None = None,
        auto_commit: bool | None = None,
        auto_expunge: bool | None = None,
        auto_refresh: bool | None = None,
        with_for_update: bool = False,
        is_scalar: bool = True,
    ) -> Sequence[SelectT] | Sequence[Any] | Sequence[Row[Any]]:
        res = self._execute(
            db_session=db_session,
            statement=statement,
            loader_options=loader_options,
            execution_options=execution_options,
            with_for_update=with_for_update,
            is_scalar=is_scalar,
            auto_commit=False,
        )
        db_objects = res.all()
        self._maybe_commit_or_flush_or_refresh_or_expunge(
            db_session=db_session,
            db_object=db_objects,
            auto_commit=auto_commit,
            auto_expunge=auto_expunge,
            auto_refresh=auto_refresh,
        )
        return db_objects

    @overload
    def execute_returning_one(
        self,
        db_session: "Session",
        statement: Select[SelectT] | Update | Delete,
        *,
        loader_options: tuple[_AbstractLoad] | None,
        execution_options: dict[str, Any] | None,
        auto_commit: bool | None,
        auto_expunge: bool | None,
        auto_refresh: bool | None,
        with_for_update: bool,
        is_scalar: Literal[True] = True,
    ) -> Any: ...

    @overload
    def execute_returning_one(
        self,
        db_session: "Session",
        statement: Select[SelectT],
        *,
        loader_options: tuple[_AbstractLoad] | None,
        execution_options: dict[str, Any] | None,
        auto_commit: bool | None,
        auto_expunge: bool | None,
        auto_refresh: bool | None,
        with_for_update: bool,
        is_scalar: Literal[False],
    ) -> Row[SelectT]: ...

    @overload
    def execute_returning_one(
        self,
        db_session: "Session",
        statement: Update | Delete,
        *,
        loader_options: tuple[_AbstractLoad] | None,
        execution_options: dict[str, Any] | None,
        auto_commit: bool | None,
        auto_expunge: bool | None,
        auto_refresh: bool | None,
        with_for_update: bool,
        is_scalar: Literal[False],
    ) -> Row[Any]: ...

    def execute_returning_one(
        self,
        db_session: "Session",
        statement: Select[SelectT] | Update | Delete,
        *,
        loader_options: tuple[_AbstractLoad] | None = None,
        execution_options: dict[str, Any] | None = None,
        auto_commit: bool | None = None,
        auto_expunge: bool | None = None,
        auto_refresh: bool | None = None,
        with_for_update: bool = False,
        is_scalar: bool = True,
    ) -> Any | Row[SelectT] | Row[Any]:
        res = self._execute(
            db_session=db_session,
            statement=statement,
            loader_options=loader_options,
            execution_options=execution_options,
            with_for_update=with_for_update,
            is_scalar=is_scalar,
            auto_commit=False,
        )
        db_object = res.one()
        self._maybe_commit_or_flush_or_refresh_or_expunge(
            db_session=db_session,
            db_object=db_object,
            auto_commit=auto_commit,
            auto_expunge=auto_expunge,
            auto_refresh=auto_refresh,
        )
        return db_object

    @overload
    def execute_returning_one_or_none(
        self,
        db_session: "Session",
        statement: Select[SelectT] | Update | Delete,
        *,
        loader_options: tuple[_AbstractLoad] | None,
        execution_options: dict[str, Any] | None,
        auto_commit: bool | None,
        auto_expunge: bool | None,
        auto_refresh: bool | None,
        with_for_update: bool,
        is_scalar: Literal[True] = True,
    ) -> Any | None: ...

    @overload
    def execute_returning_one_or_none(
        self,
        db_session: "Session",
        statement: Select[SelectT],
        *,
        loader_options: tuple[_AbstractLoad] | None,
        execution_options: dict[str, Any] | None,
        auto_commit: bool | None,
        auto_expunge: bool | None,
        auto_refresh: bool | None,
        with_for_update: bool = False,
        is_scalar: Literal[False],
    ) -> Row[SelectT] | None: ...

    @overload
    def execute_returning_one_or_none(
        self,
        db_session: "Session",
        statement: Update | Delete,
        *,
        loader_options: tuple[_AbstractLoad] | None,
        execution_options: dict[str, Any] | None,
        auto_commit: bool | None,
        auto_expunge: bool | None,
        auto_refresh: bool | None,
        with_for_update: bool = False,
        is_scalar: Literal[False],
    ) -> Row[Any] | None: ...

    def execute_returning_one_or_none(
        self,
        db_session: "Session",
        statement: Select[SelectT] | Update | Delete,
        *,
        loader_options: tuple[_AbstractLoad] | None = None,
        execution_options: dict[str, Any] | None = None,
        auto_commit: bool | None = None,
        auto_expunge: bool | None = None,
        auto_refresh: bool | None = None,
        with_for_update: bool = False,
        is_scalar: bool = True,
    ) -> Row[SelectT] | Row[Any] | Any | None:
        res = self._execute(
            db_session=db_session,
            statement=statement,
            loader_options=loader_options,
            execution_options=execution_options,
            with_for_update=with_for_update,
            is_scalar=is_scalar,
            auto_commit=False,
        )
        db_object = res.one_or_none()
        self._maybe_commit_or_flush_or_refresh_or_expunge(
            db_session=db_session,
            db_object=db_object,
            auto_commit=auto_commit,
            auto_expunge=auto_expunge,
            auto_refresh=auto_refresh,
        )
        return db_object

    def add_pagination(
        self, statement: Select[SelectT], filters: InputSchema | None
    ) -> Select[SelectT]:
        if isinstance(filters, PaginationFilter):
            statement = statement.offset(filters.offset).limit(filters.limit)

        return statement

    @overload
    def add_filters(
        self, statement: Select[SelectT], filters: InputSchema | None
    ) -> Select[SelectT]: ...

    @overload
    def add_filters(self, statement: Update, filters: InputSchema | None) -> Update: ...

    @overload
    def add_filters(self, statement: Delete, filters: InputSchema | None) -> Delete: ...

    def add_filters(
        self, statement: Select[SelectT] | Update | Delete, filters: InputSchema | None
    ) -> Select[SelectT] | Update | Delete:
        if filters is not None:
            for item, value in unpack_schema(
                schema=filters, exclude=PaginationFields
            ).items():
                if hasattr(self.model, item):
                    if isinstance(value, list):
                        statement = statement.where(
                            getattr(self.model, item).in_(value)
                        )
                    else:
                        statement = statement.where(getattr(self.model, item) == value)

        return statement

    def handle_read_only(
        self,
        is_read_only: bool | None = None,
    ) -> None:
        """Handle read-only mode for the repository.

        This method checks if the repository is in read-only mode and raises an error if necessary.
        This is to be called on editing operations like create, update, or delete.
        """
        if is_read_only is True or (is_read_only is None and self.is_read_only):
            raise MixemyRepositoryReadOnlyError(
                repository=self,
                model=self.model,
            )

    @overload
    def _prepare_statement(
        self,
        statement: Select[SelectT],
        *,
        loader_options: tuple[_AbstractLoad] | None,
        execution_options: dict[str, Any] | None,
        with_for_update: bool,
    ) -> Select[SelectT]: ...

    @overload
    def _prepare_statement(
        self,
        statement: Update,
        *,
        loader_options: tuple[_AbstractLoad] | None,
        execution_options: dict[str, Any] | None,
        with_for_update: bool,
    ) -> Update: ...

    @overload
    def _prepare_statement(
        self,
        statement: Delete,
        *,
        loader_options: tuple[_AbstractLoad] | None,
        execution_options: dict[str, Any] | None,
        with_for_update: bool,
    ) -> Delete: ...

    def _prepare_statement(
        self,
        statement: Select[SelectT] | Update | Delete,
        *,
        loader_options: tuple[_AbstractLoad] | None = None,
        execution_options: dict[str, Any] | None = None,
        with_for_update: bool = False,
    ) -> Select[SelectT] | Update | Delete:
        current_loader_options = (
            loader_options if loader_options is not None else self.loader_options
        )
        if current_loader_options is not None:
            statement = statement.options(*current_loader_options)

        current_execution_options = (
            execution_options
            if execution_options is not None
            else self.execution_options
        )
        if current_execution_options is not None:
            statement = statement.execution_options(**current_execution_options)

        if with_for_update and isinstance(statement, Select):
            statement = statement.with_for_update()

        return statement

    @staticmethod
    def _update_db_object(db_object: BaseModelT, object_in: InputSchema) -> None:
        for field, value in unpack_schema(schema=object_in).items():
            if hasattr(db_object, field):
                setattr(db_object, field, value)

    def _verify_init(self) -> None:
        for field in ["model"]:
            if not hasattr(self, field):
                raise MixemyRepositorySetupError(repository=self, undefined_field=field)


class PermissionSyncRepository(BaseSyncRepository[BaseModelT], ABC):
    """PermissionSyncRepository is an abstract base class that provides synchronous methods for handling database operations with user permission checks.

    Attributes:
        user_id_attribute (str): The attribute representing the user ID.
        user_joined_table (str | None): The attribute representing the joined table for user permissions.
        default_raise_permission_error (bool): Default flag to raise permission error.
    Methods:
        __init__(self, *, loader_options: tuple[_AbstractLoad] | None = None, execution_options: dict[str, Any] | None = None, auto_expunge: bool | None = False, auto_refresh: bool | None = True, auto_commit: bool | None = False, raise_permission_error: bool | None = True) -> None:
            Initializes the repository with the given options.
        read_with_permission(self, db_session: "Session", id: Any, user_id: Any, *, loader_options: tuple[_AbstractLoad] | None = None, execution_options: dict[str, Any] | None = None, auto_expunge: bool | None = None, auto_commit: bool | None = False, with_for_update: bool = False, raise_permission_error: bool | None = False) -> BaseModelT | None:
            Reads a single database object with permission checks.
        read_multiple_with_permission(self, db_session: "Session", user_id: Any, filters: InputSchema | None = None, *, loader_options: tuple[_AbstractLoad] | None = None, execution_options: dict[str, Any] | None = None, auto_commit: bool | None = False, auto_expunge: bool | None = None, with_for_update: bool = False) -> Sequence[BaseModelT]:
            Reads multiple database objects with permission checks.
        update_with_permission(self, db_session: "Session", id: Any, object_in: InputSchema, user_id: Any, *, loader_options: tuple[_AbstractLoad] | None = None, execution_options: dict[str, Any] | None = None, auto_commit: bool | None = None, auto_expunge: bool | None = None, auto_refresh: bool | None = None, with_for_update: bool = True, raise_permission_error: bool | None = None) -> BaseModelT | None:
            Updates a database object with permission checks.
        update_db_object_with_permission(self, db_session: "Session", db_object: BaseModelT | None, object_in: InputSchema, user_id: Any, *, auto_commit: bool | None = None, auto_expunge: bool | None = None, auto_refresh: bool | None = None, raise_permission_error: bool | None = None) -> BaseModelT | None:
            Updates a database object with permission checks.
        delete_with_permission(self, db_session: "Session", id: Any, user_id: Any, *, loader_options: tuple[_AbstractLoad] | None = None, execution_options: dict[str, Any] | None = None, auto_commit: bool | None = None, auto_expunge: bool | None = None, with_for_update: bool = False, raise_permission_error: bool | None = None) -> None:
            Deletes a database object with permission checks.
        delete_db_object_with_permission(self, db_session: "Session", db_object: BaseModelT | None, user_id: Any, *, auto_commit: bool | None = None, auto_expunge: bool | None = None, raise_permission_error: bool | None = None) -> None:
            Deletes a database object with permission checks.
        count_with_permission(self, db_session: "Session", user_id: Any, filters: InputSchema | None = None, *, loader_options: tuple[_AbstractLoad] | None = None, execution_options: dict[str, Any] | None = None, auto_commit: bool | None = False) -> int:
            Counts the number of database objects with permission checks.
        add_permission_filter(self, statement: Select[SelectT], user_id: Any) -> Select[SelectT]:
            Adds a permission filter to the SQL statement.
        handle_permission_on_db_oject(self, db_object: BaseModelT | None, user_id: Any, raise_permission_error: bool | None = None) -> BaseModelT | None:
            Checks if the user has permission on the database object.
        _verify_init(self) -> None:
            Verifies the initialization of the repository.
    """

    user_id_attribute: str = "user_id"
    user_joined_table: str | None = None

    default_raise_permission_error: bool = True
    default_permission_strategy: permission_strategies = "strict"

    def __init__(
        self,
        *,
        loader_options: tuple[_AbstractLoad] | None = None,
        execution_options: dict[str, Any] | None = None,
        auto_expunge: bool | None = None,
        auto_refresh: bool | None = None,
        auto_commit: bool | None = None,
        is_read_only: bool | None = None,
        raise_permission_error: bool | None = None,
        permission_strategy: permission_strategies | None = None,
    ) -> None:
        self.raise_permission_error = (
            raise_permission_error
            if raise_permission_error is not None
            else self.default_raise_permission_error
        )
        super().__init__(
            loader_options=loader_options,
            execution_options=execution_options,
            auto_expunge=auto_expunge,
            auto_refresh=auto_refresh,
            auto_commit=auto_commit,
            is_read_only=is_read_only,
        )
        self.permission_strategy = PermissionStrategyFactory.create_permission_strategy(
            strategy=(
                permission_strategy
                if permission_strategy is not None
                else self.default_permission_strategy
            ),
            model=self.model,
            user_id_attribute=self.user_id_attribute,
            user_joined_table=self.user_joined_table,
        )

    def read_with_permission(
        self,
        db_session: "Session",
        id: Any,
        user_id: Any,
        *,
        loader_options: tuple[_AbstractLoad] | None = None,
        execution_options: dict[str, Any] | None = None,
        auto_expunge: bool | None = None,
        auto_commit: bool | None = False,
        with_for_update: bool = False,
        raise_permission_error: bool | None = False,
    ) -> BaseModelT | None:
        db_object = self._get(
            db_session=db_session,
            id=id,
            loader_options=loader_options,
            execution_options=execution_options,
            auto_expunge=auto_expunge,
            auto_commit=auto_commit,
            with_for_update=with_for_update,
        )

        if not self.handle_permission_on_db_oject(
            db_object=db_object,
            user_id=user_id,
            raise_permission_error=raise_permission_error,
        ):
            return None

        return db_object

    def read_multiple_with_permission(
        self,
        db_session: "Session",
        user_id: Any,
        filters: InputSchema | None = None,
        *,
        loader_options: tuple[_AbstractLoad] | None = None,
        execution_options: dict[str, Any] | None = None,
        auto_commit: bool | None = False,
        auto_expunge: bool | None = None,
        with_for_update: bool = False,
    ) -> Sequence[BaseModelT]:
        statement = select(self.model)
        statement = self.add_filters(statement=statement, filters=filters)
        statement = self.add_pagination(statement=statement, filters=filters)
        statement = self.add_permission_filter(statement=statement, user_id=user_id)
        return self.execute_returning_all(
            db_session=db_session,
            statement=statement,
            loader_options=loader_options,
            execution_options=execution_options,
            auto_commit=auto_commit,
            auto_expunge=auto_expunge,
            auto_refresh=False,
            with_for_update=with_for_update,
        )

    def update_with_permission(
        self,
        db_session: "Session",
        id: Any,
        object_in: InputSchema,
        user_id: Any,
        *,
        loader_options: tuple[_AbstractLoad] | None = None,
        execution_options: dict[str, Any] | None = None,
        auto_commit: bool | None = None,
        auto_expunge: bool | None = None,
        auto_refresh: bool | None = None,
        with_for_update: bool = True,
        is_read_only: bool | None = None,
        raise_permission_error: bool | None = None,
    ) -> BaseModelT | None:
        db_object = self._get(
            db_session=db_session,
            id=id,
            loader_options=loader_options,
            execution_options=execution_options,
            auto_expunge=False,
            with_for_update=with_for_update,
            auto_commit=False,
        )

        return self.update_db_object_with_permission(
            db_session=db_session,
            user_id=user_id,
            db_object=db_object,
            object_in=object_in,
            auto_commit=auto_commit,
            auto_expunge=auto_expunge,
            auto_refresh=auto_refresh,
            is_read_only=is_read_only,
            raise_permission_error=raise_permission_error,
        )

    def update_db_object_with_permission(
        self,
        db_session: "Session",
        db_object: BaseModelT | None,
        object_in: InputSchema,
        user_id: Any,
        *,
        auto_commit: bool | None = None,
        auto_expunge: bool | None = None,
        auto_refresh: bool | None = None,
        is_read_only: bool | None = None,
        raise_permission_error: bool | None = None,
    ) -> BaseModelT | None:
        if not self.handle_permission_on_db_oject(
            db_object=db_object,
            user_id=user_id,
            raise_permission_error=raise_permission_error,
        ):
            return None

        return self.update_db_object(
            db_session=db_session,
            db_object=db_object,
            object_in=object_in,
            auto_commit=auto_commit,
            auto_expunge=auto_expunge,
            auto_refresh=auto_refresh,
            is_read_only=is_read_only,
        )

    def delete_with_permission(
        self,
        db_session: "Session",
        id: Any,
        user_id: Any,
        *,
        loader_options: tuple[_AbstractLoad] | None = None,
        execution_options: dict[str, Any] | None = None,
        auto_commit: bool | None = None,
        auto_expunge: bool | None = None,
        with_for_update: bool = False,
        is_read_only: bool | None = None,
        raise_permission_error: bool | None = None,
    ) -> None:
        db_object = self._get(
            db_session=db_session,
            id=id,
            loader_options=loader_options,
            execution_options=execution_options,
            auto_expunge=False,
            auto_commit=False,
            with_for_update=with_for_update,
        )

        self.delete_db_object_with_permission(
            db_session=db_session,
            db_object=db_object,
            user_id=user_id,
            auto_commit=auto_commit,
            auto_expunge=auto_expunge,
            is_read_only=is_read_only,
            raise_permission_error=raise_permission_error,
        )

    def delete_db_object_with_permission(
        self,
        db_session: "Session",
        db_object: BaseModelT | None,
        user_id: Any,
        *,
        auto_commit: bool | None = None,
        auto_expunge: bool | None = None,
        is_read_only: bool | None = None,
        raise_permission_error: bool | None = None,
    ) -> None:
        if not self.handle_permission_on_db_oject(
            db_object=db_object,
            user_id=user_id,
            raise_permission_error=raise_permission_error,
        ):
            return None

        return self.delete_db_object(
            db_session=db_session,
            db_object=db_object,
            auto_commit=auto_commit,
            auto_expunge=auto_expunge,
            is_read_only=is_read_only,
        )

    def count_with_permission(
        self,
        db_session: "Session",
        user_id: Any,
        filters: InputSchema | None = None,
        *,
        loader_options: tuple[_AbstractLoad] | None = None,
        execution_options: dict[str, Any] | None = None,
        auto_commit: bool | None = False,
    ) -> int:
        statement = select(func.count()).select_from(self.model)
        statement = self.add_filters(statement=statement, filters=filters)
        statement = self.add_permission_filter(statement=statement, user_id=user_id)
        return self.execute_returning_one(
            db_session=db_session,
            statement=statement,
            loader_options=loader_options,
            execution_options=execution_options,
            auto_commit=auto_commit,
            auto_expunge=False,
            auto_refresh=False,
            with_for_update=False,
        )

    def add_permission_filter(
        self,
        statement: Select[SelectT],
        user_id: Any,
    ) -> Select[SelectT]:
        return self.permission_strategy.add_permission_filter(
            statement=statement,
            user_id=user_id,
        )

    def handle_permission_on_db_oject(
        self,
        db_object: BaseModelT | None,
        user_id: Any,
        raise_permission_error: bool | None = None,
    ) -> bool:
        """Check if the user has permission on the database object.

        If the user does not have permission, it raises a MixemyRepositoryPermissionError or return false.
        If the user has permission, it returns true.
        """
        authenticated, object_id = (
            self.permission_strategy.check_permission_on_db_oject(
                db_object=db_object,
                user_id=user_id,
            )
        )

        if authenticated:
            return True

        if raise_permission_error is False or (
            raise_permission_error is None and self.raise_permission_error is False
        ):
            return False

        raise MixemyRepositoryPermissionError(
            repository=self, object_id=object_id, user_id=user_id
        )

    def _verify_init(self) -> None:
        for field in ["model", "user_id_attribute"]:
            if not hasattr(self, field):
                raise MixemyRepositorySetupError(repository=self, undefined_field=field)

        if self.user_joined_table is None and not hasattr(
            self.model, self.user_id_attribute
        ):
            raise MixemyRepositorySetupError(
                repository=self,
                undefined_field=str(self.user_id_attribute),
                message=f"User ID attribute {self.user_id_attribute} not found on model",
            )

        if self.user_joined_table is not None:
            if not hasattr(self.model, self.user_joined_table):
                raise MixemyRepositorySetupError(
                    repository=self,
                    undefined_field=str(self.user_joined_table),
                    message=f"User joined table {self.user_joined_table} not found on model",
                )
            if not hasattr(
                getattr(self.model, self.user_joined_table), self.user_id_attribute
            ):
                raise MixemyRepositorySetupError(
                    repository=self,
                    undefined_field=str(self.user_id_attribute),
                    message=f"User ID attribute {self.user_id_attribute} not found on joined table",
                )
